#include <ctype.h>

#include "postgres.h"
#include "fmgr.h"

#ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif

#include "commands/defrem.h"
#include "tsearch/dicts/spell.h"
#include "tsearch/ts_locale.h"
#include "tsearch/ts_public.h"
#include "tsearch/ts_utils.h"
#include "utils/builtins.h"
#include "utils/memutils.h"

#define SYNBUFLEN	4096
#define SYNMAXWORDS	20

typedef struct
{
	char *key; /* Word */
	char *value; /* Unparsed list of synonyms, including the word itself */
}	Syn;

typedef struct
{
	int len;
	Syn *syn;

	int keeporig;
}	DictSyn;

PG_FUNCTION_INFO_V1(dxsyn_init);
Datum dxsyn_init(PG_FUNCTION_ARGS);

PG_FUNCTION_INFO_V1(dxsyn_lexize);
Datum dxsyn_lexize(PG_FUNCTION_ARGS);

static char *
find_word(char *in, char **end)
{
	char *start;

	*end = NULL;
	while (*in && isspace(*in))
		in++;

	if (!in || *in == '#')
		return NULL;
	start = in;

	while (*in && !isspace(*in))
		in++;

	*end = in;

	return start;
}

static int
compare_syn(const void *a, const void *b)
{
	return strcmp(((Syn *) a)->key, ((Syn *) b)->key);
}

static void
read_dictionary(DictSyn *d, char *filename)
{
	FILE *fin = fopen(filename, "r");
	char buf[SYNBUFLEN];
	int cur = 0;
        
	if (!fin)
		ereport(ERROR,
				(errcode_for_file_access(),
				 errmsg("could not open file \"%s\": %m",
						filename)));
        
	while (fgets(buf, SYNBUFLEN, fin)){
		char *key;
		char *value;
		char *end = NULL;
		int slen = strlen(buf) - 1;
                
		buf[slen] = '\0';

		if (*buf == '\0')
			continue;

		if (cur == d->len){
			d->len = (d->len) ? 2 * d->len : 16;
			d->syn = (Syn *) realloc(d->syn, sizeof(Syn) * d->len);
			if (!d->syn){
				fclose(fin);
				ereport(ERROR,
						(errcode(ERRCODE_OUT_OF_MEMORY),
						 errmsg("out of memory")));
			}
		}

		key = find_word(buf, &end);
		if (!key)
			continue;

		value = buf;

		d->syn[cur].value = lowerstr(value);
		if (!d->syn[cur].value){
			fclose(fin);
			ereport(ERROR,
					(errcode(ERRCODE_OUT_OF_MEMORY),
					 errmsg("out of memory")));
		}

		d->syn[cur].key = strndup(key, end - key);

		if (!d->syn[cur].key){
			fclose(fin);
			ereport(ERROR,
					(errcode(ERRCODE_OUT_OF_MEMORY),
					 errmsg("out of memory")));
		}

		cur ++;
	}

	fclose(fin);

	d->len = cur;
	if (cur > 1)
		qsort(d->syn, d->len, sizeof(Syn), compare_syn);
}

Datum
dxsyn_init(PG_FUNCTION_ARGS)
{
	DictSyn *d = (DictSyn *) palloc(sizeof(DictSyn));
	List *dictoptions = (List *) PG_GETARG_POINTER(0);
	ListCell *l;

	d->len = 0;
	d->syn = NULL;
	d->keeporig = true;

	foreach(l, dictoptions){
		DefElem *defel = (DefElem *) lfirst(l);
		
		if (pg_strcasecmp(defel->defname, "KEEPORIG") == 0) {
			if (pg_strcasecmp(defGetString(defel), "TRUE") == 0) {
				d->keeporig = true;
			} else if (pg_strcasecmp(defGetString(defel), "FALSE") == 0) {
				d->keeporig = false;
			} else {
				elog(ERROR,"Unknown value: %s => %s", defel->defname, defGetString(defel));
			}
		} else if (pg_strcasecmp(defel->defname, "RULES") == 0) {
			read_dictionary(d, defGetString(defel));
		} else {
			elog(ERROR,"Unknown option: %s => %s",  defel->defname, defGetString(defel));
		}
	}

	MemoryContextDeleteChildren(CurrentMemoryContext);
	PG_RETURN_POINTER(d);
}

Datum
dxsyn_lexize(PG_FUNCTION_ARGS)
{
	DictSyn *d = (DictSyn *) PG_GETARG_POINTER(0);
	char *in = (char *) PG_GETARG_POINTER(1);
	int length = PG_GETARG_INT32(2);
	Syn word;
	Syn *found;
	TSLexeme *res = NULL;
        
	if (!length)
		PG_RETURN_POINTER(NULL);

	/* Create search pattern */
	{
		char *temp = pnstrdup(in, length);

		word.key = lowerstr(temp);
		pfree(temp);
		word.value = NULL;
	}

	/* Look for matching syn */
	found = (Syn *)bsearch(&word, d->syn, d->len, sizeof(Syn), compare_syn);
	pfree(word.key);

	if (!found)
		PG_RETURN_POINTER(NULL);

	/* Parse string of synonyms and return array of words */
	{
		char *value = pstrdup(found->value);
		char *pos = value;
		int value_length = strlen(value);
		int nsyns = 0;
		int is_first = true;
                
		res = palloc(0);

		while(pos < value + value_length){
			char *end;
			char *syn = find_word(pos, &end);

			if(syn){
				res = repalloc(res, sizeof(TSLexeme)*(nsyns + 2));
				res[nsyns].lexeme = NULL;
				*end = '\0';
                                
				/* Add to results if the KEEPORIG is set, or words after first one */
				if(d->keeporig || !is_first){
					res[nsyns].lexeme = pstrdup(syn);
					res[nsyns + 1].lexeme = NULL;
                                        
					nsyns ++;
				}

				is_first = false;
                                
				pos = end + 1;
			} else
				break;
		}

		pfree(value);
	}

	PG_RETURN_POINTER(res);
}
